from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from metafold.projects import Access, Project
from urllib.parse import parse_qs, urlparse
import json
import pytest

default_dt = datetime(2024, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

# Default sort order is descending by id
project_list = [
    {
        "id": "3",
        "user": "1",
        "name": "Foo",
        "access": "public",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
        "thumbnail": None,
    },
    {
        "id": "2",
        "user": "1",
        "name": "My Project",
        "access": "private",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
        "thumbnail": None,
    },
    {
        "id": "1",
        "user": "1",
        "name": "My Project",
        "access": "private",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
        "thumbnail": None,
    },
]

new_project = {
    "id": "1",
    "user": "1",
    "name": "New Project",
    "access": "public",
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
    "thumbnail": None,
}


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        u = urlparse(self.path)
        params = parse_qs(u.query)
        if u.path == "/projects":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = project_list
            if params.get("sort") == ["id:1"]:
                payload = sorted(project_list, key=lambda p: p["id"])
            elif params.get("q") == ['name:"My Project"']:
                payload = [p for p in project_list if p["name"] == "My Project"]
            self.wfile.write(json.dumps(payload).encode())
        elif u.path == "/projects/1":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(project_list[-1])
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_POST(self):
        params = self._parse_json()
        if self.path == "/projects":
            self.send_response(HTTPStatus.CREATED)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(new_project)
            if "source" in params:
                payload["id"] = "2"
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_PATCH(self):
        if self.path == "/projects/1":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(new_project).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_DELETE(self):
        if self.path == "/projects/1":
            self.send_response(HTTPStatus.OK)
            self.end_headers()
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def _parse_json(self):
        length = self.headers.get("Content-Length")
        content = self.rfile.read(int(length))
        return json.loads(content)


@pytest.fixture(scope="module")
def request_handler():
    return MockRequestHandler


def test_list_projects(client):
    projects = client.projects.list()
    assert [p.id for p in projects] == ["3", "2", "1"]


def test_list_projects_sorted(client):
    projects = client.projects.list(sort="id:1")
    assert [p.id for p in projects] == ["1", "2", "3"]


def test_list_projects_filtered(client):
    projects = client.projects.list(q='name:"My Project"')
    assert all([p.name == "My Project" for p in projects])


def test_get_project(client):
    p = client.projects.get("1")
    assert p == Project(
        id="1",
        user="1",
        name="My Project",
        access="private",
        created=default_dt,
        modified=default_dt,
    )


def test_create_project(client):
    p = client.projects.create("New Project", access=Access.PUBLIC)
    assert p == Project(
        id="1",
        user="1",
        name="New Project",
        access="public",
        created=default_dt,
        modified=default_dt,
    )


def test_duplicate_project(client):
    p = client.projects.duplicate("1", "New Project", access=Access.PUBLIC)
    assert p == Project(
        id="2",
        user="1",
        name="New Project",
        access="public",
        created=default_dt,
        modified=default_dt,
    )


def test_update_project(client):
    p = client.projects.update("1", name="New Project", access=Access.PUBLIC)
    assert p == Project(
        id="1",
        user="1",
        name="New Project",
        access="public",
        created=default_dt,
        modified=default_dt,
    )


def test_delete_project(client):
    # FIXME: Assert something
    client.projects.delete("1")



from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from metafold.assets import Asset
from metafold.jobs import Job, IO
from urllib.parse import parse_qs, urlparse
import json
import pytest

default_dt = datetime(2024, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

default_params = {
    "foo": "...",
    "bar": "...",
}

asset_json = {
    "id": "1",
    "filename": "f763df409e79eb1c.bin",
    "size": 16777216,
    "checksum": "sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "modified": "Mon, 01 Jan 2024 00:00:00 GMT",
    "project_id": "1",
}

asset_obj = Asset(
    id="1",
    filename="f763df409e79eb1c.bin",
    size=16777216,
    checksum="sha256:6310a5951d58eb3e0fdd8c8767c606615552899e65019cb1582508a7c7bfec39",
    created=default_dt,
    modified=default_dt,
    project_id="1",
)

# Default sort order is descending by id
job_list = [
    {
        "id": "3",
        "name": "bar",
        "type": "evaluate_graph",
        "state": "success",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "error": None,
        "inputs": {
            "params": default_params,
        },
        "outputs": {
            "params": None,
        },
        "needs": [],
        "project_id": "1",
        "workflow_id": None,
        "parameters": default_params,
        "meta": None,
    },
    {
        "id": "2",
        "name": "foo",
        "type": "evaluate_graph",
        "state": "success",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "error": None,
        "inputs": {
            "params": default_params,
        },
        "outputs": {
            "params": None,
        },
        "needs": [],
        "project_id": "1",
        "workflow_id": None,
        "parameters": default_params,
        "meta": None,
    },
    {
        "id": "1",
        "name": "foo",
        "type": "evaluate_graph",
        "state": "success",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "error": None,
        "inputs": {
            "params": default_params,
        },
        "outputs": {
            "params": None,
        },
        "needs": [],
        "project_id": "1",
        "workflow_id": None,
        "parameters": default_params,
        "meta": None,
    },
]

new_job = {
    "id": "1",
    "name": "My Job",
    "type": "test_job",
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "started": "Mon, 01 Jan 2024 00:00:00 GMT",
    "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
    "error": None,
    "inputs": {
        "params": {
            "foo": "1",
            "bar": "a",
            "baz": "[2, \"b\"]",
        },
    },
    "outputs": {
        "params": None,
    },
    "needs": [],
    "project_id": "1",
    "workflow_id": None,
    "assets": [],
    "parameters": {
        "foo": "1",
        "bar": "a",
        "baz": "[2, \"b\"]",
    },
    "meta": None,
}

poll_count: int = 0


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        u = urlparse(self.path)
        params = parse_qs(u.query)
        if u.path == "/projects/1/jobs":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = job_list
            if params.get("sort") == ["id:1"]:
                payload = sorted(job_list, key=lambda p: p["id"])
            elif params.get("q") == ["name:foo"]:
                payload = [p for p in job_list if p["name"] == "foo"]
            self.wfile.write(json.dumps(payload).encode())
        elif u.path == "/projects/1/jobs/1":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(job_list[-1])
            payload["assets"] = [asset_json]
            self.wfile.write(json.dumps(payload).encode())
        elif u.path == "/projects/1/jobs/1/status":
            global poll_count
            poll_count += 1
            if poll_count < 3:
                self.send_response(HTTPStatus.ACCEPTED)
                payload = deepcopy(new_job)
                payload["state"] = "started"
            else:
                self.send_response(HTTPStatus.CREATED)
                payload = deepcopy(new_job)
                payload.update({
                    "state": "success",
                    "assets": [asset_json],
                })
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_POST(self):
        if self.path == "/projects/1/jobs":
            self.send_response(HTTPStatus.ACCEPTED)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(new_job)
            payload.update({
                "state": "pending",
                "link": "http://localhost:8000/projects/1/jobs/1/status",
            })
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_PATCH(self):
        if self.path == "/projects/1/jobs/1":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(job_list[-1])
            payload.update({
                "name": "baz",
                "assets": [asset_json],
            })
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)


@pytest.fixture(scope="module")
def request_handler():
    return MockRequestHandler


def test_list_jobs(client):
    jobs = client.jobs.list()
    assert [j.id for j in jobs] == ["3", "2", "1"]


def test_list_jobs_sorted(client):
    jobs = client.jobs.list(sort="id:1")
    assert [j.id for j in jobs] == ["1", "2", "3"]


def test_list_jobs_filtered(client):
    jobs = client.jobs.list(q="name:foo")
    assert all([j.name == "foo" for j in jobs])


def test_get_job(client):
    j = client.jobs.get("1")
    assert j == Job(
        id="1",
        name="foo",
        type="evaluate_graph",
        state="success",
        created=default_dt,
        started=default_dt,
        finished=default_dt,
        error=None,
        inputs=IO(params=default_params),
        outputs=IO(),
        needs=[],
        project_id="1",
        workflow_id=None,
        assets=[asset_obj],
        parameters=default_params,
        meta=None,
    )


def test_run_job(client):
    params = {
        "foo": "1",
        "bar": "a",
        "baz": "[2, \"b\"]",
    }
    j = client.jobs.run("test_job", params, name="My Job")
    assert j == Job(
        id="1",
        name="My Job",
        type="test_job",
        state="success",
        created=default_dt,
        started=default_dt,
        finished=default_dt,
        inputs=IO(params=params),
        outputs=IO(),
        needs=[],
        project_id="1",
        workflow_id=None,
        assets=[asset_obj],
        parameters=params,
        meta=None,
    )


def test_poll_job(client):
    params = {
        "foo": "1",
        "bar": "a",
        "baz": "[2, \"b\"]",
    }
    url = client.jobs.run_status("test_job", params, name="My Job")
    assert url == "http://localhost:8000/projects/1/jobs/1/status"

    r = client.jobs.poll(url)
    assert Job(**r.json()) == Job(
        id="1",
        name="My Job",
        type="test_job",
        state="success",
        created=default_dt,
        started=default_dt,
        finished=default_dt,
        inputs=IO(params=params),
        outputs=IO(),
        needs=[],
        project_id="1",
        workflow_id=None,
        assets=[asset_obj],
        parameters=params,
        meta=None,
    )


def test_update_job(client):
    j = client.jobs.update("1", name="baz")
    assert j.name == "baz"



from attrs import field, frozen
from datetime import datetime
from metafold.api import asdatetime, asdict, optional_datetime
from metafold.client import Client
from metafold.exceptions import PollTimeout
from requests import Response
from typing import Optional, Union


@frozen(kw_only=True)
class Workflow:
    """Workflow resource.

    Attributes:
        id: Workflow ID.
        state: Workflow state. May be one of: pending, started, success, failure, or
            canceled.
        created: Workflow creation datetime.
        started: Workflow started datetime.
        finished: Workflow finished datetime.
        definition: Workflow definition string.
        project_id: Project ID.
    """
    id: str
    jobs: list[str] = field(factory=list)
    state: str
    created: datetime = field(converter=asdatetime)
    started: Optional[datetime] = field(
        converter=lambda v: optional_datetime(v), default=None)
    finished: Optional[datetime] = field(
        converter=lambda v: optional_datetime(v), default=None)
    definition: str
    project_id: str


class WorkflowsEndpoint:
    """Metafold workflows endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(
        self,
        sort: Optional[str] = None,
        q: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> list[Workflow]:
        """List jobs.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "created", "started", or "finished".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id" and "state".
            project_id: Workflow project ID.

        Returns:
            List of job resources.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/workflows"
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get(url, params=payload)
        return [Workflow(**w) for w in r.json()]

    def get(self, workflow_id: str, project_id: Optional[str] = None) -> Workflow:
        """Get a workflow.

        Args:
            workflow_id: ID of workflow to get.
            project_id: Workflow project ID.

        Returns:
            Workflow resource.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/workflows/{workflow_id}"
        r: Response = self._client.get(url)
        return Workflow(**r.json())

    def run(
        self, definition: str,
        parameters: Optional[dict[str, str]] = None,
        assets: Optional[dict[str, str]] = None,
        timeout: Union[int, float] = 120,
        project_id: Optional[str] = None,
    ) -> Workflow:
        """Dispatch a new workflow and wait for it to complete.

        Workflow completion does not indicate success. Access the completed workflow's
        state to check for success/failure.

        Args:
            definition: Workflow definition YAML.
            parameters: Parameter mapping for jobs in the definition.
            assets: Asset mapping for jobs in the definition.
            timeout: Time in seconds to wait for a result.
            project_id: Workflow project ID.

        Returns:
            Completed workflow resource.
        """
        project_id = self._client.project_id(project_id)
        payload = asdict(definition=definition, parameters=parameters, assets=assets)
        r: Response = self._client.post(f"/projects/{project_id}/workflows", json=payload)
        url = r.json()["link"]
        try:
            r = self._client.poll(url, timeout)
        except PollTimeout as e:
            raise RuntimeError(
                f"Workflow failed to complete within {timeout} seconds"
            ) from e
        return Workflow(**r.json())

    def cancel(self, workflow_id: str, project_id: Optional[str] = None) -> Workflow:
        """Cancel a running workflow.

        Args:
            workflow_id: ID of workflow to cancel.
            project_id: Workflow project ID.

        Returns:
            Workflow resource.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/workflows/{workflow_id}/cancel"
        r: Response = self._client.post(url)
        return Workflow(**r.json())

    def delete(self, workflow_id: str, project_id: Optional[str] = None):
        """Delete a workflow.

        Args:
            workflow_id: ID of workflow to delete.
            project_id: Workflow project ID.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/workflows/{workflow_id}"
        self._client.delete(url)



from attrs import field, frozen
from datetime import datetime
from metafold.api import asdatetime, asdict, optional, optional_datetime
from metafold.assets import Asset
from metafold.client import Client
from metafold.exceptions import PollTimeout
from requests import Response
from typing import Any, Optional, TypedDict, Union
from typing_extensions import TypeAlias


def _assets(v: list[Union[dict[str, Any], Asset]]) -> list[Asset]:
    return [a if isinstance(a, Asset) else Asset(**a) for a in v]


AssetDict: TypeAlias = dict[str, Union[dict[str, Any], Asset]]


def _assets_dict(v: AssetDict) -> dict[str, Asset]:
    return dict(
        (k, a if isinstance(a, Asset) else Asset(**a))
        for k, a in v.items()
    )


class IODict(TypedDict):
    params: Optional[dict[str, Any]]
    assets: Optional[dict[str, AssetDict]]


@frozen(kw_only=True)
class IO:
    """Job input/output.

    Attributes:
        params: JSON-encoded parameter values.
        assets: Related assets.
    """
    params: Optional[dict[str, Any]] = None
    assets: Optional[dict[str, Asset]] = field(
        converter=lambda v: optional(_assets_dict)(v), default=None)

    @staticmethod
    def from_dict(d: IODict) -> "IO":
        return IO(params=d.get("params"), assets=d.get("assets"))


@frozen(kw_only=True)
class Job:
    """Job resource.

    Attributes:
        id: Job ID.
        name: Job name.
        type: Job type.
        state: Job state. May be one of: pending, started, success, failure, or
            canceled.
        created: Job creation datetime.
        started: Job started datetime.
        finished: Job finished datetime.
        error: Error message for failed jobs.
        inputs: Input assets and parameters.
        outputs: Output assets and parameters.
        needs: List of upstream job IDs in a workflow graph.
        project_id: Project ID.
        workflow_id: Workflow ID.
        assets: (Deprecated) List of generated asset resources.
        parameters: (Deprecated) Job parameters.
        meta: (Deprecated) Additional metadata generated by the job.
    """
    id: str
    name: Optional[str] = None
    type: str
    state: str
    created: datetime = field(converter=asdatetime)
    started: Optional[datetime] = field(
        converter=lambda v: optional_datetime(v), default=None)
    finished: Optional[datetime] = field(
        converter=lambda v: optional_datetime(v), default=None)
    error: Optional[str] = None
    inputs: IO = field(converter=lambda v: v if isinstance(v, IO) else IO.from_dict(v))
    outputs: IO = field(converter=lambda v: v if isinstance(v, IO) else IO.from_dict(v))
    needs: list[str]
    project_id: Optional[str] = None
    workflow_id: Optional[str] = None
    # NOTE(ryan): Deprecated
    assets: Optional[list[Asset]] = field(
        converter=lambda v: optional(_assets)(v), default=None)
    parameters: dict[str, Any]
    meta: dict[str, Any]


class JobsEndpoint:
    """Metafold jobs endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(
        self,
        sort: Optional[str] = None,
        q: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> list[Job]:
        """List jobs.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "name", "created", "started", or
                "finished".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id", "name", "type", and "state".
            project_id: Job project ID.

        Returns:
            List of job resources.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/jobs"
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get(url, params=payload)
        return [Job(**j) for j in r.json()]

    def get(self, job_id: str, project_id: Optional[str] = None) -> Job:
        """Get a job.

        Args:
            job_id: ID of job to get.
            project_id: Job project ID.

        Returns:
            Job resource.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/jobs/{job_id}"
        r: Response = self._client.get(url)
        return Job(**r.json())

    def run(
        self, type: str, params: dict[str, Any],
        name: Optional[str] = None,
        timeout: Union[int, float] = 120,
        project_id: Optional[str] = None,
    ) -> Job:
        """Dispatch a new job and wait for a result.

        See Metafold API docs for the full list of jobs.

        Args:
            type: Job type.
            params: Job parameters.
            name: Optional job name.
            timeout: Time in seconds to wait for a result.
            project_id: Job project ID.

        Returns:
            Completed job resource.
        """
        url = self.run_status(type, params, name=name, project_id=project_id)
        try:
            r: Response = self.poll(url, timeout)
        except PollTimeout as e:
            raise RuntimeError(
                f"Job '{name or type}' failed to complete within {timeout} seconds"
            ) from e
        return Job(**r.json())

    def run_status(
        self, type: str, params: dict[str, Any],
        name: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> str:
        """Dispatch a new job and return immediately without waiting for result.

        See Metafold API docs for the full list of jobs.

        Args:
            type: Job type.
            params: Job parameters.
            name: Optional job name.
            project_id: Job project ID.

        Returns:
            Job status url.
        """
        project_id = self._client.project_id(project_id)
        payload = asdict(type=type, parameters=params, name=name)
        r: Response = self._client.post(f"/projects/{project_id}/jobs", json=payload)
        return r.json()["link"]

    def poll(self, *args, **kwargs):
        return self._client.poll(*args, **kwargs)

    def update(
        self, job_id: str,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> Job:
        """Update a job.

        Args:
            job_id: ID of job to update.
            name: New job name. The existing name remains unchanged if None.
            project_id: Job project ID.

        Returns:
            Updated job resource.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/jobs/{job_id}"
        payload = asdict(name=name)
        r: Response = self._client.patch(url, data=payload)
        return Job(**r.json())

    def delete(self, job_id: str, project_id: Optional[str] = None):
        """Delete a job.

        Args:
            job_id: ID of job to delete.
            project_id: Job project ID.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/jobs/{job_id}"
        self._client.delete(url)



from attrs import field, frozen
from datetime import datetime
from enum import Enum
from metafold.api import asdatetime, asdict
from metafold.client import Client
from metafold.func_types import Graph
from requests import Response
from typing import Any, Optional, Union


class Access(Enum):
    """Project access scope.

    Attributes:
        PRIVATE: Project is private to owner.
        PUBLIC: Project may be accessed by unauthenticated users.
    """
    PRIVATE = "private"
    PUBLIC = "public"


def check_access(a: str) -> Access:
    if a.upper() not in Access.__dict__:
        raise ValueError("Expected 'private' or 'public'")
    return Access[a.upper()]


@frozen(kw_only=True)
class Project:
    """Project resource.

    Attributes:
        id: Project ID.
        user: Project user ID.
        name: Project name.
        access: Project access.
        created: Project creation datetime.
        modified: Project last modified datetime.
        thumbnail: URL to project thumbnail.
        project: Arbitrary project data.
        graph: Project graph data.
    """
    id: str
    user: str
    name: str
    access: Access = field(converter=check_access)
    created: datetime = field(converter=asdatetime)
    modified: datetime = field(converter=asdatetime)
    thumbnail: Optional[str] = None
    project: Optional[dict[str, Any]] = None
    graph: Optional[Graph] = None


class ProjectsEndpoint:
    """Metafold projects endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(self, sort: Optional[str] = None, q: Optional[str] = None) -> list[Project]:
        """List projects.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "user", "name", "created", and
                "modified".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id", "user", and "name".

        Returns:
            List of project resources.
        """
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get("/projects", params=payload)
        return [Project(**a) for a in r.json()]

    def get(self, id: Optional[str] = None) -> Project:
        """Get an project.

        Args:
            id: Override ID of project to get. Defaults to client project ID.

        Returns:
            Project resource.
        """
        id = self._client.project_id(id)
        url = f"/projects/{id}"
        r: Response = self._client.get(url)
        return Project(**r.json())

    def create(
        self, name: str,
        access: Union[Access, str] = Access.PRIVATE,
        data: Optional[dict[str, Any]] = None,
    ) -> Project:
        """Create a project.

        Args:
            name: Project name.
            access: Project access. By default projects are private.
            data: Optional project data. This parameter accepts arbitrary
                JSON-serializable data. Helpful for tracking application state.

        Returns:
            Project resource.
        """
        if isinstance(access, str):
            access = check_access(access)
        payload = asdict(name=name, access=access.value, project=data)
        r: Response = self._client.post("/projects", json=payload)
        return Project(**r.json())

    def duplicate(
        self, id: str, name: str,
        access: Union[Access, str] = Access.PRIVATE,
    ) -> Project:
        """Duplicate a project.

        Args:
            id: Project to duplicate.
            name: New project name.
            access: New project access. By default projects are private.

        Returns:
            Project resource.
        """
        if isinstance(access, str):
            access = check_access(access)
        payload = asdict(source=id, name=name, access=access.value)
        r: Response = self._client.post("/projects", json=payload)
        return Project(**r.json())

    def update(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        access: Optional[Union[Access, str]] = None,
        data: Optional[dict[str, Any]] = None,
        graph: Optional[Graph] = None,
    ) -> Project:
        """Update a project.

        Args:
            id: Override ID of project to update. Defaults to client project ID.
            name: Optional project name.
            access: Optional project access.
            data: Optional project data. This parameter accepts arbitrary
                JSON-serializable data. Helpful for tracking application state.
            graph: Optional shape JSON.

        Returns:
            Updated project resource.
        """
        if access and isinstance(access, str):
            access = check_access(access)
        id = self._client.project_id(id)
        url = f"/projects/{id}"
        payload = asdict(name=name, project=data, graph=graph)
        # Access is optional, but we also need to cast to str
        if isinstance(access, Access):
            payload["access"] = access.value
        r: Response = self._client.patch(url, json=payload)
        return Project(**r.json())

    def delete(self, id: str) -> None:
        """Delete an project.

        Args:
            id: Override ID of project to update. Defaults to client project ID.
        """
        id = self._client.project_id(id)
        url = f"/projects/{id}"
        self._client.delete(url)



from metafold.auth import AuthProvider
from metafold.exceptions import PollTimeout
from requests import HTTPError, Response, Session
from typing import Any, Callable, Optional, Union
from urllib.parse import urljoin
import platform
import time


class Client:
    """Base client."""

    def __init__(
        self,
        base_url: str,
        access_token: Optional[str] = None,
        project_id: Optional[str] = None,
        auth: Optional[AuthProvider] = None
    ) -> None:
        if bool(auth) == bool(access_token):
            raise ValueError(
                "Expected AuthProvider or access_token to be provided"
            )
        self._auth = auth
        self._default_project = project_id
        self._base_url = base_url
        self._session = Session()
        self._session.headers.update({
            "Accept": "application/json",
            "User-Agent": f"Python/{platform.python_version()}",
        })
        if access_token:
            self._session.headers.update({
                "Authorization": f"Bearer {access_token}",
            })

    def project_id(self, id: Optional[str] = None) -> str:
        id = id or self._default_project
        if not id:
            raise ValueError(
                "Project ID required, set a default ID when initializing the client"
            )
        return id

    def set_project_id(self, id: str) -> None:
        self._default_project = id

    def _request(
        self, request: Callable[..., Response], url: str,
        *args: Any, **kwargs: Any,
    ) -> Response:
        url = urljoin(self._base_url, url)
        headers = None
        if self._auth:
            headers = {"Authorization": f"Bearer {self._auth.get_token()}"}
        r: Response = request(url, *args, **kwargs, headers=headers)
        if not r.ok:
            body: dict[str, Any] = r.json()
            # Error responses aren't entirely consistent in the Metafold API,
            # for now we check for a handful of possible fields.
            reason = body.get("errors") or body.get("msg") or body.get("description")
            raise HTTPError(f"HTTP error occurred: {reason or r.reason}")
        return r

    def get(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.get, url, *args, **kwargs)

    def post(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.post, url, *args, **kwargs)

    def put(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.put, url, *args, **kwargs)

    def patch(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.patch, url, *args, **kwargs)

    def delete(self, url: str, *args: Any, **kwargs: Any) ->  Response:
        return self._request(self._session.delete, url, *args, **kwargs)

    def poll(
        self, url: str,
        timeout: Union[int, float] = 120,
        every: Union[int, float] = 1,
        ) -> Response:
        """Poll the given URL in regular intervals.

        Helpful for waiting on async processes given a status URL.

        Args:
            timeout: Time in seconds to wait for a result.
            every: Frequency in seconds.

        Returns:
            HTTP response.
        """
        t0 = time.monotonic()
        r = self.get(url)
        while r.status_code == 202:
            elapsed = time.monotonic() - t0
            if elapsed >= timeout:
                raise PollTimeout(f"Polling timed out: {url}")
            time.sleep(1)
            r = self.get(url)
        return r



from abc import abstractmethod
from enum import Enum
from typing import (
    Any,
    Annotated,
    Generic,
    Literal,
    Optional,
    Protocol,
    TypeVar,
    TypedDict,
    Union,
)
from typing import cast
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

try:
    import numpy as np
    _USE_NUMPY = True
except ImportError:
    _USE_NUMPY = False


Vec2i: TypeAlias = Union[
    Annotated[list[int], 2],
    "np.ndarray[tuple[Literal[2]], np.dtype[np.int32]]",
]
Vec2f: TypeAlias = Union[
    Annotated[list[float], 2],
    "np.ndarray[tuple[Literal[2]], np.dtype[np.float32]]",
]
Vec3i: TypeAlias = Union[
    Annotated[list[int], 3],
    "np.ndarray[tuple[Literal[3]], np.dtype[np.int32]]",
]
Vec3f: TypeAlias = Union[
    Annotated[list[float], 3],
    "np.ndarray[tuple[Literal[3]], np.dtype[np.float32]]",
]
Vec4i: TypeAlias = Union[
    Annotated[list[int], 4],
    "np.ndarray[tuple[Literal[4]], np.dtype[np.int32]]",
]
Vec4f: TypeAlias = Union[
    Annotated[list[float], 4],
    "np.ndarray[tuple[Literal[4]], np.dtype[np.float32]]",
]
Mat2f: TypeAlias = Union[
    Annotated[list[float], 4],
    "np.ndarray[tuple[Literal[2], Literal[2]], np.dtype[np.float32]]",
]
Mat3f: TypeAlias = Union[
    Annotated[list[float], 9],
    "np.ndarray[tuple[Literal[3], Literal[3]], np.dtype[np.float32]]",
]
Mat4f: TypeAlias = Union[
    Annotated[list[float], 16],
    "np.ndarray[tuple[Literal[4], Literal[4]], np.dtype[np.float32]]",
]


class UnitCell(TypedDict):
    nodes: Union[list[Vec3f], "np.ndarray"]
    edges: Union[list[Vec2i], "np.ndarray"]


class Asset(TypedDict):
    path: str


class VolumeAsset(Asset):
    file_type: Literal["Raw"]


class TriangleMeshAsset(Asset):
    file_type: Literal["MeshFile"]


class CustomShapeAsset(Asset):
    file_type: Literal["ShaderBinarySPIRV"]


class ParametrizationAsset(Asset): ...
class LineNetworkAsset(Asset): ...
class LineNetworkBvhAsset(Asset): ...
class MeshBvhAsset(Asset): ...


class FuncType(Enum):
    BYTE  = "Byte"
    INT   = "Int"
    FLOAT = "Float"
    VEC2I = "Vec2i"
    VEC2F = "Vec2f"
    VEC3I = "Vec3i"
    VEC3F = "Vec3f"
    VEC4I = "Vec4i"
    VEC4F = "Vec4f"
    MAT2F = "Mat2f"
    MAT3F = "Mat3f"
    MAT4F = "Mat4f"


T = TypeVar("T", bound=FuncType, covariant=True)


class Result: ...
class TypedResult(Result, Generic[T]): ...


Inputs: TypeAlias = dict[str, Result]
Assets: TypeAlias = dict[str, Asset]
Params: TypeAlias = dict[str, Any]


class Evaluator(Protocol):
    def __call__(
        self,
        type_: str,
        inputs: Optional[Inputs] = None,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> Result:
        ...


POINT_SOURCE_VAR_TYPE = "_PointSource"


class Func(Protocol):
    def __call__(self, eval_: Evaluator) -> Result: ...


class TypedFunc(Func, Protocol[T]): ...


class BaseEvaluator(Evaluator):
    def __init__(self, source: Optional[TypedFunc[Literal[FuncType.VEC3F]]] = None):
        self._source = source

    def __call__(
        self,
        type_: str,
        inputs: Optional[Inputs] = None,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> Result:
        if type_ == POINT_SOURCE_VAR_TYPE:
            if not self._source:
                raise Exception("Expected valid point source")
            return self._source(self)
        return self._eval(type_, inputs, assets, parameters)

    @abstractmethod
    def _eval(
        self,
        type_: str,
        inputs: Optional[Inputs] = None,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> Result:
        ...


class OperatorParams(TypedDict, total=False):
    parameters: dict[str, Any]


class Operator(OperatorParams):
    type: str


class Edge(TypedDict):
    source: int
    target: Annotated[list[Union[int, str]], 2]


class Graph(TypedDict):
    operators: list[Operator]
    edges: list[Edge]


class IndexResult(TypedResult):
    def __init__(self, index: int):
        self.index = index


class JSONEvaluator(BaseEvaluator):
    def __init__(self, source: Optional[TypedFunc[Literal[FuncType.VEC3F]]] = None):
        super().__init__(source=source)
        self._operators: list[Operator] = []
        self._edges: list[Edge] = []

    def json(self):
        return {
            "operators": self._operators,
            "edges": self._edges,
        }

    def _eval(
        self,
        type_: str,
        inputs: Optional[Inputs] = None,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> IndexResult:
        operator: Operator = self._make_operator(
            type_, assets=assets, parameters=parameters)
        target = len(self._operators)
        self._operators.append(operator)
        if inputs:
            for name, result in inputs.items():
                self._edges.append({
                    "source": cast(IndexResult, result).index,
                    "target": [target, name],
                })
        return IndexResult(target)

    def _make_operator(
        self,
        type_: str,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> Operator:
        operator: Operator = {"type": type_}
        if assets:
            parameters = parameters or {}
            parameters.update(assets)
        if parameters:
            for k, v in parameters.items():
                if _USE_NUMPY and isinstance(v, np.ndarray):
                    parameters[k] = v.flatten().tolist()
            operator["parameters"] = parameters
        return operator



from datetime import datetime, timezone
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, Union


def asdatetime(s: Union[str, datetime]) -> datetime:
    """Parse Metafold API datetime.

    Note datetime strings returned by the Metafold API are RFC 1123 formatted,
    times are always in GMT.

    Returns:
        UTC datetime.
    """
    if isinstance(s, datetime):
        return s
    return datetime.strptime(s, "%a, %d %b %Y %H:%M:%S %Z").replace(tzinfo=timezone.utc)


def asdict(**kwargs: Any) -> dict[str, Any]:
    """Convert present kwargs to dictionary."""
    d = {}
    for k, v in kwargs.items():
        if v is not None:
            d[k] = v
    return d


T = TypeVar("T")
U = TypeVar("U")


def optional(f: Callable[[T], U]) -> Callable[[Optional[T]], Optional[U]]:
    """Decorator to generate converters that accept optional values."""
    @wraps(f)
    def decorator(v: Optional[T]) -> Optional[U]:
        if v is None:
            return None
        return f(v)

    return decorator


optional_datetime = optional(asdatetime)



from metafold.client import Client
from metafold.projects import ProjectsEndpoint
from metafold.assets import AssetsEndpoint
from metafold.jobs import JobsEndpoint
from metafold.workflows import WorkflowsEndpoint
from metafold.auth import AuthProvider
from typing import Optional


class MetafoldClient(Client):
    """Metafold REST API client.

    Attributes:
        projects: Sub-client for projects endpoint.
        assets: Sub-client for assets endpoint.
        jobs: Sub-client for jobs endpoint.
    """
    projects: ProjectsEndpoint
    assets: AssetsEndpoint
    jobs: JobsEndpoint
    workflows: WorkflowsEndpoint

    def __init__(
        self,
        access_token: Optional[str] = None,
        project_id: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        auth_domain: str = "metafold3d.us.auth0.com",
        base_url: str = "https://api.metafold3d.com/",
    ) -> None:
        """Initialize Metafold API client.

        Args:
            access_token: Metafold API secret key.
            project_id: ID of the project to make API calls against.
            base_url: Metafold API URL. Used for internal testing.
        """
        # client_id and client_secret have priority
        if not any([client_id and client_secret, access_token]):
            raise ValueError(
                "Expected client_id and client_secret or access_token to be provided"
            )
        elif client_id and client_secret:
            auth = AuthProvider(client_id, client_secret, auth_domain, base_url)
            super().__init__(base_url, auth=auth, project_id=project_id)
        else:
            super().__init__(base_url, access_token=access_token, project_id=project_id)

        self.projects = ProjectsEndpoint(self)
        self.assets = AssetsEndpoint(self)
        self.jobs = JobsEndpoint(self)
        self.workflows = WorkflowsEndpoint(self)



import numpy as np
from metafold.func_types import JSONEvaluator
from metafold.func import (
    CSGIntersect,
    EllipsoidPrimitive,
    GenerateSamplePoints,
    LoadVolume,
    PointSource,
    Redistance,
    SampleSurfaceLattice,
    SampleSurfaceLattice_Parameters,
    SampleVolume,
    Threshold,
    VolumeAsset,
)
from metafold.nx import GraphEvaluator


def test_build_graph_json():
    source = GenerateSamplePoints({
        "size": [2, 2, 2],
        "resolution": [64, 64, 64],
    })

    gyroidParams: SampleSurfaceLattice_Parameters = {
        "lattice_type": "Gyroid",
        "scale": [0.25, 0.25, 0.25],
    }


    # NOTE: This should fail to compile during type checking
    # f = Redistance(CSGIntersect(source, EllipsoidPrimitive(source)))

    want = {
        "operators": [
            {
                "type": "GenerateSamplePoints",
                "parameters": {
                    "size": [2, 2, 2],
                    "resolution": [64, 64, 64],
                },
            },
            {
                "type": "SampleSurfaceLattice",
                "parameters": {
                    "lattice_type": "Gyroid",
                    "scale": [0.25, 0.25, 0.25],
                },
            },
            {
                "type": "SampleBox",
                "parameters": {
                    "shape_type": "Ellipsoid",
                    "size": [1, 1, 1],
                },
            },
            {
                "type": "CSG",
                "parameters": {
                    "operation": "Intersect",
                },
            },
            {"type": "Redistance"},
            {"type": "Threshold"},
        ],
        "edges": [
            {"source": 0, "target": [1, "Points"]},   # GenerateSamplePoints -> SampleBox
            {"source": 0, "target": [2, "Points"]},   # GenerateSamplePoints -> SampleSurfaceLattice
            {"source": 1, "target": [3, "A"]},        # SampleSurfaceLattice -> CSG
            {"source": 2, "target": [3, "B"]},        #            SampleBox -> CSG
            {"source": 3, "target": [4, "Samples"]},  #                  CSG -> Redistance
            {"source": 4, "target": [5, "Samples"]},  #           Redistance -> Threshold
        ],
    }

    f = Threshold(
        Redistance(
            CSGIntersect(
                SampleSurfaceLattice(source, gyroidParams),
                EllipsoidPrimitive(source, {"size": [1.0, 1.0, 1.0]}),
            ),
        ),
    )
    e = JSONEvaluator()
    f(e)

    assert e.json() == want

    # With deferred point source
    f = Threshold(
        Redistance(
            CSGIntersect(
                SampleSurfaceLattice(PointSource, gyroidParams),
                EllipsoidPrimitive(PointSource, {"size": [1.0, 1.0, 1.0]}),
            ),
        ),
    )
    e = JSONEvaluator(source)
    f(e)

    assert e.json() == want


def test_build_graph_json_with_assets():
    source = GenerateSamplePoints({
        "size": [2, 2, 2],
        "resolution": [64, 64, 64],
    })

    volumeAsset: VolumeAsset = {
        "file_type": "Raw",
        "path": "foo.bin",
    }
    volume = LoadVolume(volumeAsset, {"resolution": [64, 64, 64]})

    f = Threshold(
        Redistance(
            SampleVolume(source, volume, {
                "volume_size": [2, 2, 2],
                "volume_offset": [-1, -1, -1],
            }),
        ),
    )
    e = JSONEvaluator()
    f(e)

    assert e.json() == {
        "operators": [
            {
                "type": "GenerateSamplePoints",
                "parameters": {
                    "size": [2, 2, 2],
                    "resolution": [64, 64, 64],
                },
            },
            {
                "type": "LoadVolume",
                "parameters": {
                    "volume_data": volumeAsset,
                    "resolution": [64, 64, 64],
                },
            },
            {
                "type": "SampleVolume",
                "parameters": {
                    "volume_size": [2, 2, 2],
                    "volume_offset": [-1, -1, -1],
                },
            },
            {"type": "Redistance"},
            {"type": "Threshold"},
        ],
        "edges": [
            {"source": 0, "target": [2, "Points"]},   # GenerateSamplePoints -> SampleVolume
            {"source": 1, "target": [2, "Volume"]},   #           LoadVolume -> SampleVolume
            {"source": 2, "target": [3, "Samples"]},  #         SampleVolume -> Redistance
            {"source": 3, "target": [4, "Samples"]},  #           Redistance -> Threshold
        ],
    }


def test_build_nx_graph():
    source = GenerateSamplePoints({
        "size": [2, 2, 2],
        "resolution": [64, 64, 64],
    })

    gyroidParams: SampleSurfaceLattice_Parameters = {
        "lattice_type": "Gyroid",
        "scale": [0.25, 0.25, 0.25],
    }

    edges = [
        # (source, target, key)
        ("GenerateSamplePoints", "SampleSurfaceLattice", "Points"),
        ("GenerateSamplePoints", "SampleBox", "Points"),
        ("SampleSurfaceLattice", "CSG", "A"),
        ("SampleBox", "CSG", "B"),
        ("CSG", "Redistance", "Samples"),
        ("Redistance", "Threshold", "Samples"),
    ]

    f = Threshold(
        Redistance(
            CSGIntersect(
                SampleSurfaceLattice(source, gyroidParams),
                EllipsoidPrimitive(source, {"size": [1.0, 1.0, 1.0]}),
            ),
        ),
    )
    e = GraphEvaluator()
    f(e)

    want = set(edges)
    for source_operator, target_operator, key in e.graph.edges:
        want.remove((source_operator.type, target_operator.type, key))
    assert len(want) == 0, f"{want} edges not found in graph"

    # With deferred point source
    f = Threshold(
        Redistance(
            CSGIntersect(
                SampleSurfaceLattice(PointSource, gyroidParams),
                EllipsoidPrimitive(PointSource, {"size": [1.0, 1.0, 1.0]}),
            ),
        ),
    )
    e = GraphEvaluator(source)
    f(e)

    want = set(edges)
    for source_operator, target_operator, key in e.graph.edges:
        want.remove((source_operator.type, target_operator.type, key))
    assert len(want) == 0, f"{want} edges not found in graph"



from auth0.authentication import GetToken  # type: ignore
from collections import namedtuple
from datetime import datetime, timedelta, timezone
from typing import Optional

Token = namedtuple("Token", ["access_token", "expires_at"])


class AuthProvider:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        auth_domain: str,
        base_url: str
    ) ->  None:
        self._auth_domain = auth_domain
        self._base_url = base_url
        self._client_id = client_id

        self._get_token = GetToken(
            self._auth_domain,
            self._client_id,
            client_secret=client_secret,
        )
        self._token: Optional[Token] = None

    def get_token(self) -> str:
        now = datetime.now(timezone.utc)
        if not self._token or self._token.expires_at - now < timedelta(minutes=1):
            token = self._get_token.client_credentials(self._base_url)
            expires_at = now + timedelta(seconds=token["expires_in"])
            self._token = Token(token["access_token"], expires_at)
        return self._token.access_token



from dataclasses import dataclass
from metafold.func_types import (
    Assets,
    BaseEvaluator,
    FuncType,
    Inputs,
    Params,
    TypedFunc,
    TypedResult,
)
from typing import Literal, Optional, cast
import networkx as nx


@dataclass(eq=False)
class Operator(object):
    type: str
    assets: Optional[Assets]
    params: Optional[Params]


class OperatorResult(TypedResult):
    def __init__(self, operator: Operator):
        self.operator = operator


class GraphEvaluator(BaseEvaluator):
    def __init__(self, source: Optional[TypedFunc[Literal[FuncType.VEC3F]]] = None):
        super().__init__(source=source)
        self._g: nx.MultiDiGraph[Operator] = nx.MultiDiGraph()

    @property
    def graph(self):
        return self._g

    def _eval(
        self,
        type_: str,
        inputs: Optional[Inputs] = None,
        assets: Optional[Assets] = None,
        parameters: Optional[Params] = None,
    ) -> OperatorResult:
        operator = Operator(type_, assets, parameters)
        self._g.add_node(operator)
        if inputs:
            for name, result in inputs.items():
                source = cast(OperatorResult, result).operator
                self._g.add_edge(source, operator, key=name)
        return OperatorResult(operator)



from attrs import field, frozen
from datetime import datetime
from metafold.api import asdatetime, asdict
from metafold.client import Client
from os import PathLike
from requests import Response
from typing import IO, Optional, Union
import requests


@frozen(kw_only=True)
class Asset:
    """Asset resource.

    Attributes:
        id: Asset ID.
        filename: Asset filename.
        size: File size in bytes.
        checksum: File checksum.
        created: Asset creation datetime.
        modified: Asset last modified datetime.
        project_id: Project ID.
        job_id: Job ID.
    """
    id: str
    filename: str
    size: int
    checksum: str
    created: datetime = field(converter=asdatetime)
    modified: datetime = field(converter=asdatetime)
    project_id: str
    job_id: Optional[str] = None


class AssetsEndpoint:
    """Metafold assets endpoint."""

    def __init__(self, client: Client) -> None:
        self._client = client

    def list(
        self,
        sort: Optional[str] = None,
        q: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> list[Asset]:
        """List assets.

        Args:
            sort: Sort string. For details on syntax see the Metafold API docs.
                Supported sorting fields are: "id", "filename", "size", "created", or
                "modified".
            q: Query string. For details on syntax see the Metafold API docs.
                Supported search fields are: "id" and "filename".
            project_id: Asset project ID.

        Returns:
            List of asset resources.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/assets"
        payload = asdict(sort=sort, q=q)
        r: Response = self._client.get(url, params=payload)
        return [Asset(**a) for a in r.json()]

    def get(self, asset_id: str, project_id: Optional[str] = None) -> Asset:
        """Get an asset.

        Args:
            asset_id: ID of asset to get.
            project_id: Asset project ID.

        Returns:
            Asset resource.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/assets/{asset_id}"
        r: Response = self._client.get(url)
        return Asset(**r.json())

    def download_file(
        self, asset_id: str, path: Union[str, PathLike],
        project_id: Optional[str] = None,
    ):
        """Download an asset.

        Args:
            asset_id: ID of asset to download.
            path: Path to downloaded file.
            project_id: Asset project ID.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/assets/{asset_id}"
        r: Response = self._client.get(url, params={"download": "true"})
        r = requests.get(r.json()["link"], stream=True)
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=65536):  # 64 KiB
                f.write(chunk)

    def create(
        self, f: Union[str, bytes, PathLike, IO[bytes]],
        project_id: Optional[str] = None,
    ) -> Asset:
        """Upload an asset.

        Args:
            f: File-like object (opened in binary mode) or path to file on disk.
            project_id: Asset project ID.

        Returns:
            Asset resource.
        """
        project_id = self._client.project_id(project_id)
        fp: IO[bytes] = _open_file(f)
        try:
            url = f"/projects/{project_id}/assets"
            r: Response = self._client.post(url, files={"file": fp})
        finally:
            fp.close()
        return Asset(**r.json())

    def delete(self, asset_id: str, project_id: Optional[str] = None) -> None:
        """Delete an asset.

        Args:
            asset_id: ID of asset to delete.
            project_id: Asset project ID.
        """
        project_id = self._client.project_id(project_id)
        url = f"/projects/{project_id}/assets/{asset_id}"
        self._client.delete(url)


def _open_file(f: Union[str, bytes, PathLike, IO[bytes]]) -> IO[bytes]:
    if isinstance(f, (str, bytes, PathLike)):
        return open(f, "rb")
    return f



# ruff: noqa: F401
# This file was automatically generated from the source file: func.py.in.
# Any edits should be made to the template file before re-running the codegen.
from typing import Literal, Optional, TypedDict, TypeVar, TYPE_CHECKING, Union
from typing import cast
import sys
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

try:
    import numpy as np
except ImportError:
    pass

from .func_types import (
    Evaluator,
    Func,
    FuncType,
    Assets,
    Inputs,
    POINT_SOURCE_VAR_TYPE,
    Params,
    Result,
    TypedFunc,
    TypedResult,

    # Parameter types
    Vec2i,
    Vec2f,
    Vec3i,
    Vec3f,
    Vec4i,
    Vec4f,
    Mat2f,
    Mat3f,
    Mat4f,
    UnitCell,

    # Asset types
    VolumeAsset,
    TriangleMeshAsset,
    CustomShapeAsset,
    ParametrizationAsset,
    LineNetworkAsset,
    LineNetworkBvhAsset,
    MeshBvhAsset,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    F = TypeVar("F", bound=Callable)
    def cache(f: F) -> F: ...
else:
    from functools import cache


@cache
def PointSource(eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
    return cast(TypedResult[Literal[FuncType.VEC3F]], eval_(POINT_SOURCE_VAR_TYPE))


CSG_Enum_operation: TypeAlias = Literal["Intersect", "Subtract", "Union"]


class CSG_Parameters(TypedDict, total=False):
    operation: CSG_Enum_operation
    smoothing: float


class CSG(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        a: TypedFunc[Literal[FuncType.FLOAT]],
        b: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[CSG_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "A": a,
            "B": b,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "CSG",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


ComputeCurvatures_Enum_spacing_type: TypeAlias = Literal["Continuous", "Discrete"]


class ComputeCurvatures_Parameters(TypedDict, total=False):
    spacing_type: ComputeCurvatures_Enum_spacing_type
    step_size: float
    volume_size: Vec3f


class ComputeCurvatures(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        samples: Func,
        parameters: Optional[ComputeCurvatures_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Samples": samples,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "ComputeCurvatures",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class ComputeNormals_Parameters(TypedDict, total=False):
    volume_offset: Vec3f
    volume_size: Vec3f
    xform: Mat4f


class ComputeNormals(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        volume: Func,
        parameters: Optional[ComputeNormals_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
            "Volume": volume,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "ComputeNormals",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class GenerateSamplePoints_Parameters(TypedDict, total=False):
    offset: Vec3f
    resolution: Vec3i
    size: Vec3f
    xform: Mat4f


class GenerateSamplePoints(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        parameters: Optional[GenerateSamplePoints_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = None
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "GenerateSamplePoints",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


GradeCellSize_Enum_shape_type: TypeAlias = Literal["Box", "Cylinder", "Ellipsoid", "Plane"]


class GradeCellSize_Parameters(TypedDict, total=False):
    num_steps: int
    offset: float
    shape_size: Vec3f
    shape_type: GradeCellSize_Enum_shape_type
    shape_xform: Mat4f
    step_size: float
    width: float
    xform: Mat4f


class GradeCellSize(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[GradeCellSize_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "GradeCellSize",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class InterpolateBoundaryCoords_Parameters(TypedDict, total=False):
    tolerance: float
    xform: Mat4f


class InterpolateBoundaryCoords(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        boundary_data: ParametrizationAsset,
        mesh_data: TriangleMeshAsset,
        parameters: Optional[InterpolateBoundaryCoords_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = {
            "boundary_data": boundary_data,
            "mesh_data": mesh_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "InterpolateBoundaryCoords",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class LinearFilter_Parameters(TypedDict, total=False):
    scale_a: float
    scale_b: float


class LinearFilter(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        inputa: TypedFunc[Literal[FuncType.FLOAT]],
        inputb: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[LinearFilter_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "InputA": inputa,
            "InputB": inputb,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "LinearFilter",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class LoadSamplePoints_Parameters(TypedDict, total=False):
    points: Union[list[Vec3f], "np.ndarray"]


class LoadSamplePoints(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        parameters: Optional[LoadSamplePoints_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = None
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "LoadSamplePoints",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


LoadVolume_Enum_component_type: TypeAlias = Literal["Byte", "Float", "Integer", "None", "Vec2f", "Vec2i", "Vec3f", "Vec3i", "Vec4f", "Vec4i"]


class LoadVolume_Parameters(TypedDict, total=False):
    component_type: LoadVolume_Enum_component_type
    resolution: Vec3i


class LoadVolume(Func):
    def __init__(
        self,
        volume_data: VolumeAsset,
        parameters: Optional[LoadVolume_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = None
        self.assets: Optional[Assets]
        self.assets = {
            "volume_data": volume_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> Result:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "LoadVolume",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return r


MapTexturePrimitive_Enum_shape_type: TypeAlias = Literal["Box", "Cylinder", "Ellipsoid", "Plane"]


class MapTexturePrimitive_Parameters(TypedDict, total=False):
    box_width: Vec3f
    scale: Vec2f
    shape_type: MapTexturePrimitive_Enum_shape_type
    shape_xform: Mat4f
    xform: Mat4f


class MapTexturePrimitive(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        image: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[MapTexturePrimitive_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
            "Image": image,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "MapTexturePrimitive",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class Redistance_Parameters(TypedDict, total=False):
    post_offset: float
    pre_offset: float


class Redistance(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        samples: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[Redistance_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Samples": samples,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "Redistance",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


SampleBeam_Enum_node_type: TypeAlias = Literal["None", "Sphere"]


SampleBeam_Enum_section_type: TypeAlias = Literal["Box", "Circle", "Cross"]


class SampleBeam_Parameters(TypedDict, total=False):
    node_radius: float
    node_type: SampleBeam_Enum_node_type
    section_radius: float
    section_type: SampleBeam_Enum_section_type
    smoothing: float
    xform: Mat4f


class SampleBeam(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        bvh_data: LineNetworkBvhAsset,
        network_data: LineNetworkAsset,
        parameters: Optional[SampleBeam_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = {
            "bvh_data": bvh_data,
            "network_data": network_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleBeam",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


SampleBox_Enum_shape_type: TypeAlias = Literal["Box", "BoxFrame", "CappedCone", "Capsule", "Cylinder", "Ellipsoid", "Link", "Plane", "Torus"]


class SampleBox_Parameters(TypedDict, total=False):
    free_param: float
    shape_type: SampleBox_Enum_shape_type
    size: Vec3f
    xform: Mat4f


class SampleBox(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[SampleBox_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleBox",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class SampleCustomShape_Parameters(TypedDict, total=False):
    xform: Mat4f


class SampleCustomShape(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        shader_data: CustomShapeAsset,
        parameters: Optional[SampleCustomShape_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = {
            "shader_data": shader_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleCustomShape",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


SampleLattice_Enum_node_type: TypeAlias = Literal["None", "Sphere"]


SampleLattice_Enum_section_type: TypeAlias = Literal["Box", "Circle", "Cross"]


class SampleLattice_Parameters(TypedDict, total=False):
    lattice_data: UnitCell
    node_radius: float
    node_type: SampleLattice_Enum_node_type
    scale: Vec3f
    scale_grading_range: Vec2f
    scale_grading_scale: Vec2f
    section_radius: float
    section_radius_grading_range: Vec2f
    section_radius_grading_scale: Vec2f
    section_type: SampleLattice_Enum_section_type
    smoothing: float
    xform: Mat4f


class SampleLattice(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[SampleLattice_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleLattice",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class SampleSpinodoid_Parameters(TypedDict, total=False):
    angles: Vec3f
    density: float
    pore_size: float
    wave_count: int
    xform: Mat4f


class SampleSpinodoid(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[SampleSpinodoid_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleSpinodoid",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


SampleSurfaceLattice_Enum_lattice_type: TypeAlias = Literal["CD", "CI2Y", "CP", "CPM_Y", "CS", "CY", "C_Y", "D", "F", "FRD", "Gyroid", "I2Y", "IWP", "None", "P", "PM_Y", "S", "SD1", "Schwarz", "SchwarzD", "SchwarzN", "SchwarzPW", "SchwarzW", "W", "Y"]


class SampleSurfaceLattice_Parameters(TypedDict, total=False):
    lattice_type: SampleSurfaceLattice_Enum_lattice_type
    scale: Vec3f
    scale_grading_range: Vec2f
    scale_grading_scale: Vec2f
    threshold: float
    threshold_grading_offset: Vec2f
    threshold_grading_range: Vec2f
    xform: Mat4f


class SampleSurfaceLattice(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[SampleSurfaceLattice_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleSurfaceLattice",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class SampleTriangleMesh_Parameters(TypedDict, total=False):
    cw_winding: int
    xform: Mat4f


class SampleTriangleMesh(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        mesh_data: TriangleMeshAsset,
        parameters: Optional[SampleTriangleMesh_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = {
            "mesh_data": mesh_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleTriangleMesh",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class SampleTriangleMeshBvh_Parameters(TypedDict, total=False):
    xform: Mat4f


class SampleTriangleMeshBvh(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        bvh_data: MeshBvhAsset,
        mesh_data: TriangleMeshAsset,
        parameters: Optional[SampleTriangleMeshBvh_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = {
            "bvh_data": bvh_data,
            "mesh_data": mesh_data,
        }
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleTriangleMeshBvh",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class SampleVolume_Parameters(TypedDict, total=False):
    volume_offset: Vec3f
    volume_size: Vec3f
    xform: Mat4f


class SampleVolume(Func):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        volume: Func,
        parameters: Optional[SampleVolume_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
            "Volume": volume,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> Result:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "SampleVolume",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return r


class Shell_Parameters(TypedDict, total=False):
    offset: float
    thickness: float


class Shell(TypedFunc[Literal[FuncType.FLOAT]]):
    def __init__(
        self,
        samples: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[Shell_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Samples": samples,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.FLOAT]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "Shell",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.FLOAT]], r)


class Threshold_Parameters(TypedDict, total=False):
    width: float


class Threshold(TypedFunc[Literal[FuncType.BYTE]]):
    def __init__(
        self,
        samples: TypedFunc[Literal[FuncType.FLOAT]],
        parameters: Optional[Threshold_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Samples": samples,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.BYTE]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "Threshold",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.BYTE]], r)


class TransformCylindricalCoords_Parameters(TypedDict, total=False):
    radial_scale: float
    xform: Mat4f


class TransformCylindricalCoords(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[TransformCylindricalCoords_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "TransformCylindricalCoords",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class TransformMirrorCoords_Parameters(TypedDict, total=False):
    mirror_normal: Vec3f
    mirror_point: Vec3f
    xform: Mat4f


class TransformMirrorCoords(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[TransformMirrorCoords_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "TransformMirrorCoords",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


class TransformSphericalCoords_Parameters(TypedDict, total=False):
    xform: Mat4f


class TransformSphericalCoords(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[TransformSphericalCoords_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "TransformSphericalCoords",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


TransformTwistCoords_Enum_axis: TypeAlias = Literal["X", "Y", "Z"]


class TransformTwistCoords_Parameters(TypedDict, total=False):
    axis: TransformTwistCoords_Enum_axis
    frequency: float
    xform: Mat4f


class TransformTwistCoords(TypedFunc[Literal[FuncType.VEC3F]]):
    def __init__(
        self,
        points: TypedFunc[Literal[FuncType.VEC3F]],
        parameters: Optional[TransformTwistCoords_Parameters] = None,
    ):
        self.inputs: Optional[dict[str, Func]]
        self.inputs = {
            "Points": points,
        }
        self.assets: Optional[Assets]
        self.assets = None
        self.parameters = parameters

    @cache
    def __call__(self, eval_: Evaluator) -> TypedResult[Literal[FuncType.VEC3F]]:
        inputs: Optional[Inputs] = None
        if self.inputs:
            inputs = dict((k, v(eval_)) for k, v in self.inputs.items())
        r = eval_(
            "TransformTwistCoords",
            inputs=inputs,
            assets=self.assets,
            # https://github.com/python/mypy/issues/4976#issuecomment-460971843
            parameters=cast(Optional[Params], self.parameters),
        )
        return cast(TypedResult[Literal[FuncType.VEC3F]], r)


def BoxPrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Box"
    return SampleBox(points, parameters=parameters)


def BoxFramePrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "BoxFrame"
    return SampleBox(points, parameters=parameters)


def CappedConePrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "CappedCone"
    return SampleBox(points, parameters=parameters)


def CapsulePrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Capsule"
    return SampleBox(points, parameters=parameters)


def CylinderPrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Cylinder"
    return SampleBox(points, parameters=parameters)


def EllipsoidPrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Ellipsoid"
    return SampleBox(points, parameters=parameters)


def LinkPrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Link"
    return SampleBox(points, parameters=parameters)


def PlanePrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Plane"
    return SampleBox(points, parameters=parameters)


def TorusPrimitive(
    points: TypedFunc[Literal[FuncType.VEC3F]],
    parameters: Optional[SampleBox_Parameters] = None,
  ) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters = parameters or {}
    parameters["shape_type"] = "Torus"
    return SampleBox(points, parameters=parameters)


def CSGIntersect(
    a: TypedFunc[Literal[FuncType.FLOAT]],
    b: TypedFunc[Literal[FuncType.FLOAT]],
    smoothing: Optional[float] = None,
) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters: CSG_Parameters = {"operation": "Intersect"}
    if smoothing is not None:
        parameters["smoothing"] = smoothing
    return CSG(a, b, parameters=parameters)


def CSGSubtract(
    a: TypedFunc[Literal[FuncType.FLOAT]],
    b: TypedFunc[Literal[FuncType.FLOAT]],
    smoothing: Optional[float] = None,
) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters: CSG_Parameters = {"operation": "Subtract"}
    if smoothing is not None:
        parameters["smoothing"] = smoothing
    return CSG(a, b, parameters=parameters)


def CSGUnion(
    a: TypedFunc[Literal[FuncType.FLOAT]],
    b: TypedFunc[Literal[FuncType.FLOAT]],
    smoothing: Optional[float] = None,
) -> TypedFunc[Literal[FuncType.FLOAT]]:
    parameters: CSG_Parameters = {"operation": "Union"}
    if smoothing is not None:
        parameters["smoothing"] = smoothing
    return CSG(a, b, parameters=parameters)



from numpy.typing import ArrayLike
from scipy.spatial.transform import Rotation as R  # type: ignore
from typing import Optional
import numpy as np


def xform(
    translation: Optional[ArrayLike] = None,
    rotation: Optional[ArrayLike] = None,
) -> np.ndarray:
    """Compose transformation matrix.

       Args:
            translation: Translation in the x, y, and z directions.
            rotation: Euler angles in degrees. Rotation is applied in the order Rz, Ry, Rx.

        Returns:
            4 x 4 affine transformation matrix.
    """
    translation = translation or np.array([0.0, 0.0, 0.0])
    rotation = rotation or np.array([0.0, 0.0, 0.0])

    r = R.from_euler("xyz", rotation,  degrees=True).as_matrix()
    m = np.eye(4)
    m[:3, :3] = r
    m[3, :3] = translation
    return m



from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from metafold.workflows import Workflow
from urllib.parse import parse_qs, urlparse
import json
import pytest

default_dt = datetime(2024, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

# Default sort order is descending by id
workflow_list = [
    {
        "id": "3",
        "jobs": ["1", "2"],
        "state": "success",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "definition": "...",
        "project_id": "1",
    },
    {
        "id": "2",
        "jobs": ["1", "2"],
        "state": "started",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "definition": "...",
        "project_id": "1",
    },
    {
        "id": "1",
        "jobs": ["1", "2"],
        "state": "success",
        "created": "Mon, 01 Jan 2024 00:00:00 GMT",
        "started": "Mon, 01 Jan 2024 00:00:00 GMT",
        "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
        "definition": "...",
        "project_id": "1",
    },
]

new_workflow = {
    "id": "1",
    "jobs": ["1", "2"],
    "created": "Mon, 01 Jan 2024 00:00:00 GMT",
    "started": "Mon, 01 Jan 2024 00:00:00 GMT",
    "finished": "Mon, 01 Jan 2024 00:00:00 GMT",
    "definition": "foo",
    "project_id": "1",
}

poll_count: int = 0


class MockRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        u = urlparse(self.path)
        params = parse_qs(u.query)
        if u.path == "/projects/1/workflows":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = workflow_list
            if params.get("sort") == ["id:1"]:
                payload = sorted(workflow_list, key=lambda p: p["id"])
            elif params.get("q") == ["state:started"]:
                payload = [p for p in workflow_list if p["state"] == "started"]
            self.wfile.write(json.dumps(payload).encode())
        elif u.path == "/projects/1/workflows/1":
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(workflow_list[-1])
            self.wfile.write(json.dumps(payload).encode())
        elif u.path == "/projects/1/workflows/1/status":
            global poll_count
            poll_count += 1
            if poll_count < 3:
                self.send_response(HTTPStatus.ACCEPTED)
                payload = deepcopy(new_workflow)
                payload["state"] = "started"
            else:
                self.send_response(HTTPStatus.CREATED)
                payload = deepcopy(new_workflow)
                payload["state"] = "success"
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)

    def do_POST(self):
        if self.path == "/projects/1/workflows":
            self.send_response(HTTPStatus.ACCEPTED)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            payload = deepcopy(new_workflow)
            payload.update({
                "state": "pending",
                "link": "http://localhost:8000/projects/1/workflows/1/status",
            })
            self.wfile.write(json.dumps(payload).encode())
        else:
            self.send_error(HTTPStatus.NOT_FOUND)


@pytest.fixture(scope="module")
def request_handler():
    return MockRequestHandler


def test_list_workflows(client):
    workflows = client.workflows.list()
    assert [w.id for w in workflows] == ["3", "2", "1"]


def test_list_workflows_sorted(client):
    workflows = client.workflows.list(sort="id:1")
    assert [w.id for w in workflows] == ["1", "2", "3"]


def test_list_workflows_filtered(client):
    workflows = client.workflows.list(q="state:started")
    assert all([w.state == "started" for w in workflows])


def test_get_workflow(client):
    w = client.workflows.get("1")
    assert w == Workflow(
        id="1",
        jobs=["1", "2"],
        state="success",
        created=default_dt,
        started=default_dt,
        finished=default_dt,
        definition="...",
        project_id="1",
    )


def test_run_workflow(client):
    definition = "foo"
    w = client.workflows.run(definition)
    assert w == Workflow(
        id="1",
        jobs=["1", "2"],
        state="success",
        created=default_dt,
        started=default_dt,
        finished=default_dt,
        definition=definition,
        project_id="1",
    )
